from .crowdstrike import crowdstrike_fdr_pipeline
from .crowdstrike import crowdstrike_falcon_pipeline

pipelines = {
    "crowdstrike_fdr": crowdstrike_fdr_pipeline,
    "crowdstrike_falcon": crowdstrike_falcon_pipeline,
}



from sigma.conversion.state import ConversionState
from sigma.conversion.base import TextQueryBackend
from sigma.conditions import (
    ConditionItem,
    ConditionAND,
    ConditionOR,
    ConditionNOT,
    ConditionFieldEqualsValueExpression,
)
from sigma.conversion.deferred import (
    DeferredQueryExpression,
    DeferredTextQueryExpression,
)
from sigma.types import (
    SigmaCompareExpression,
    SigmaRegularExpressionFlag,
    SpecialChars,
    SigmaString,
)
from sigma.exceptions import SigmaFeatureNotSupportedByBackendError
import sigma
import re
from typing import ClassVar, Dict, Tuple, Pattern, Optional, Union


class LogScaleDeferredEqualsOperator(DeferredTextQueryExpression):
    template = "{field}{op}/{value}/i"
    operators = {
        True: "!=",
        False: "=",
    }


class LogScaleDeferredTestOperator(DeferredQueryExpression):
    template = "test({field1}=={field2})"


class LogScaleDeferredInOperator(DeferredTextQueryExpression):
    template = "{op}in({field}, ignoreCase=true, values=[{value}])"
    operators = {
        True: "!",
        False: "",
    }


class LogScaleDeferredCIDRExpression(DeferredTextQueryExpression):
    template = "{op}cidr({field}, subnet={value})"
    operators = {
        True: "!",
        False: "",
    }


class LogScaleDeferredRegularExpression(DeferredTextQueryExpression):
    template = "{op}regex(field={field},{value})"
    operators = {True: "!", False: ""}


class LogScaleBackend(TextQueryBackend):
    """CrowdStrike LogScale backend."""

    # Operator precedence: tuple of Condition{AND,OR,NOT} in order of precedence.
    # The backend generates grouping if required
    name: ClassVar[str] = "CrowdStrike LogScale backend"
    formats: Dict[str, str] = {
        "default": "CrowdStrike LogScale queries",
    }
    requires_pipeline: bool = True

    # Operator precedence: tuple of Condition{AND,OR,NOT} in order of precedence.
    precedence: ClassVar[Tuple[ConditionItem, ConditionItem, ConditionItem]] = (
        ConditionNOT,
        ConditionOR,
        ConditionAND,
    )
    group_expression: ClassVar[str] = "({expr})"

    # Generated query tokens
    token_separator: str = " "
    or_token: ClassVar[str] = "or"
    and_token: ClassVar[str] = " "
    not_token: ClassVar[str] = "not"
    eq_token: ClassVar[str] = "="

    # String output
    ## Fields
    ### Quoting
    field_quote: ClassVar[str] = '"'
    field_quote_pattern: ClassVar[Pattern] = re.compile(
        r"^[@|#]?[\w.]+$"
    )  # Some logscale fields start with # or @
    field_quote_pattern_negation: ClassVar[bool] = True

    ### Escaping
    # fields in LogScale are defined directly by CrowdStrike so this is probably unnecessary but let's do it anyway
    field_escape: ClassVar[str] = "\\"
    field_escape_quote: ClassVar[bool] = True

    ## Values
    ## This is unused by the backend because {field}={value} is case sensitive in logscale. All values are treated as regular expressions
    ## https://library.humio.com/data-analysis/writing-queries-operations.html#writing-queries-operations-strings-case-insensitive
    str_quote: ClassVar[str] = (
        '"'  # string quoting character (added as escaping character)
    )
    str_quote_pattern_negation = False
    escape_char: ClassVar[str] = (
        "\\"  # Escaping character for special characrers inside string
    )
    wildcard_multi: ClassVar[str] = ".*"  # Character used as multi-character wildcard
    wildcard_single: ClassVar[str] = "."  # Character used as single-character wildcard
    add_escaped: ClassVar[str] = (
        "\\"  # Characters quoted in addition to wildcards and string quote
    )
    filter_chars: ClassVar[str] = ""  # Characters filtered
    bool_values: ClassVar[Dict[bool, str]] = (
        {  # Values to which boolean values are mapped.
            True: "true",
            False: "false",
        }
    )

    ## Special Regex Values for LogScale
    str_quote_re: ClassVar[str] = ""
    escape_char_re: ClassVar[str] = "\\"
    wildcard_multi_re: ClassVar[str] = ".*"
    wildcard_single_re: ClassVar[str] = "."
    add_escaped_re: ClassVar[str] = "*$^.|?()[]+/{}"
    filter_chars_re: ClassVar[str] = ""
    bool_values_re: ClassVar[Dict[bool, str]] = {
        True: "true",
        False: "false",
    }

    # Regular expressions
    # Regular expression query as format string with placeholders {field}, {regex}, {flag_x} where x
    # is one of the flags shortcuts supported by Sigma (currently i, m and s) and refers to the
    # token stored in the class variable re_flags.
    re_expression: ClassVar[str] = "{field}=/{regex}/{flag_i}{flag_m}{flag_s}"
    re_exact_match: ClassVar[str] = "{field}=/^{regex}$/{flag_i}{flag_m}{flag_s}"
    re_escape_char: ClassVar[str] = "\\"
    re_escape: ClassVar[Tuple[str]] = '"'
    re_escape_escape_char: bool = False
    re_flag_prefix: bool = False
    re_flags: Dict[SigmaRegularExpressionFlag, str] = {
        SigmaRegularExpressionFlag.IGNORECASE: "i",
        SigmaRegularExpressionFlag.MULTILINE: "m",
        SigmaRegularExpressionFlag.DOTALL: "s",
    }

    # case sensitive match is just {field}={value} in logscale
    case_sensitive_match_expression: ClassVar[str] = "{field}={value}"

    # wildcards could have been used here as well but we went with the regex format without the case insensitivity flag
    case_sensitive_startswith_expression: ClassVar[str] = "{field}=/^{value}/"
    case_sensitive_endswith_expression: ClassVar[str] = "{field}=/{value}$/"
    case_sensitive_contains_expression: ClassVar[str] = "{field}=/{value}/"

    # also handled as regex. Look at the convert_condition_field_eq_val_str method
    startswith_expression: ClassVar[str] = "{field}=/^{regex}/{flag_i}{flag_m}{flag_s}"
    endswith_expression: ClassVar[str] = "{field}=/{regex}$/{flag_i}{flag_m}{flag_s}"
    contains_expression: ClassVar[str] = "{field}=/{regex}/{flag_i}{flag_m}{flag_s}"

    # https://library.humio.com/data-analysis/functions-cidr.html
    # Convert method is overloaded below
    cidr_expression: ClassVar[Optional[str]] = "{value}"

    # Numeric comparison operators
    compare_op_expression: ClassVar[str] = (
        "{field}{operator}{value}"  # Compare operation query as format string with placeholders {field}, {operator} and {value}
    )
    # https://library.humio.com/data-analysis/syntax-operators.html
    compare_operators: ClassVar[Dict[SigmaCompareExpression.CompareOperators, str]] = {
        SigmaCompareExpression.CompareOperators.LT: "<",
        SigmaCompareExpression.CompareOperators.LTE: "<=",
        SigmaCompareExpression.CompareOperators.GT: ">",
        SigmaCompareExpression.CompareOperators.GTE: ">=",
    }

    # Null/None expressions
    # https://library.humio.com/kb/kb-empty-fields.html
    field_null_expression: ClassVar[str] = (
        "{field}!=*"  # Expression for field has null value as format string with {field} placeholder for field name
    )

    # Field existence condition expressions.
    # I dont think humio can do the below
    # field_exists_expression : ClassVar[str] = "exists({field})"             # Expression for field existence as format string with {field} placeholder for field name
    # field_not_exists_expression : ClassVar[str] = "notexists({field})"      # Expression for field non-existence as format string with {field} placeholder for field name. If not set, field_exists_expression is negated with boolean NOT.

    # https://library.humio.com/data-analysis/functions-in.html """
    # Logscale does not support 'or' with in statements so we decided not to use it. However the logic is here if this changes
    convert_or_as_in: ClassVar[bool] = False  # Convert OR as in-expression
    convert_and_as_in: ClassVar[bool] = False  # Convert AND as in-expression
    in_expressions_allow_wildcards: ClassVar[bool] = (
        True  # Values in list can contain wildcards. If set to False (default) only plain values are converted into in-expressions.
    )

    field_in_list_expression: ClassVar[str] = (
        "{list}"  # Expression for field in list of values as format string with placeholders {field}, {op} and {list}
    )
    or_in_operator: ClassVar[str] = (
        "in"  # Operator used to convert OR into in-expressions. Must be set if convert_or_as_in is set
    )
    list_separator: ClassVar[str] = ", "  # List element separator

    # Value not bound to a field
    # We want these to be case insensivitive so use the regex representation again
    unbound_value_str_expression: ClassVar[str] = (
        "/{value}/i"  # Expression for string value not bound to a field as format string with placeholder {value}
    )
    unbound_value_num_expression: ClassVar[str] = (
        "/{value}/i"  # Expression for number value not bound to a field as format string with placeholder {value}
    )
    unbound_value_re_expression: ClassVar[str] = (
        "/{value}/{flag_i}{flag_m}{flag_s}"  # Expression for regular expression not bound to a field as format string with placeholder {value} and {flag_x} as described for re_expression
    )

    # Query finalization: appending and concatenating deferred query part
    deferred_start: ClassVar[str] = (
        "| "  # String used as separator between main query and deferred parts
    )
    deferred_separator: ClassVar[str] = (
        "| "  # String used to join multiple deferred query parts
    )
    deferred_only_query: ClassVar[str] = (
        ""  # String used as query if final query only contains deferred expression
    )

    def __init__(
        self,
        processing_pipeline: Optional[
            "sigma.processing.pipeline.ProcessingPipeline"
        ] = None,
        collect_errors: bool = False,
        **kwargs,
    ):
        super().__init__(processing_pipeline, collect_errors, **kwargs)

    def convert_condition_field_eq_val_cidr(
        self,
        cond: ConditionFieldEqualsValueExpression,
        state: ConversionState,
    ) -> LogScaleDeferredCIDRExpression:
        """Defer CIDR network range matching to pipelined where cidrmatch command after main search expression."""
        if cond.parent_condition_chain_contains(ConditionOR):
            raise SigmaFeatureNotSupportedByBackendError(
                "ORing CIDR matching is not yet supported by LogScale backend",
                source=cond.source,
            )
        return LogScaleDeferredCIDRExpression(
            state, cond.field, super().convert_condition_field_eq_val_cidr(cond, state)
        ).postprocess(None, cond)

    def convert_condition_field_eq_val_str(
        self, cond: ConditionFieldEqualsValueExpression, state: ConversionState
    ) -> Union[str, DeferredQueryExpression]:
        """Conversion of field = string value expressions. In logscale we handle everything as regexs"""
        try:
            if (  # contains: string starts and ends with wildcard
                self.contains_expression is not None
                and cond.value.startswith(SpecialChars.WILDCARD_MULTI)
                and cond.value.endswith(SpecialChars.WILDCARD_MULTI)
            ):
                expr = self.contains_expression
                value = cond.value[1:-1]
            elif (  # Same as above but for 'endswith' operator: string starts with wildcard
                self.endswith_expression is not None
                and cond.value.startswith(SpecialChars.WILDCARD_MULTI)
            ):
                expr = self.endswith_expression
                value = cond.value[1:]

            elif (  # Same as above but for 'startswith' operator: string ends with wildcard
                self.startswith_expression
                is not None  # 'startswith' operator is defined in backend
                and cond.value.endswith(
                    SpecialChars.WILDCARD_MULTI
                ) 
            ):
                expr = self.startswith_expression
                # If all conditions are fulfilled, use 'startswith' operator instead of equal token
                value = cond.value[:-1]
            else:
                expr = self.re_exact_match
                value = cond.value
            return expr.format(
                field=self.escape_and_quote_field(cond.field),
                regex=self.convert_value_str_re(value, state),
                flag_i="i",
                flag_m="",
                flag_s="",
            )
        except TypeError:  # pragma: no cover
            raise NotImplementedError(
                "Field equals string value expressions with strings are not supported by the backend."
            )

    def convert_value_str_re(self, s: SigmaString, state: ConversionState) -> str:
        converted = s.convert(
            escape_char=self.escape_char_re,
            wildcard_multi=self.wildcard_multi_re,
            wildcard_single=self.wildcard_single_re,
            add_escaped=self.str_quote_re + self.add_escaped_re + self.escape_char_re,
            filter_chars=self.filter_chars_re,
        )
        return converted



from .logscale import LogScaleBackend

backends = {  # Mapping between backend identifiers and classes. This is used by the pySigma plugin system to recognize backends and expose them with the identifier.
    "CrowdStrike Logscale": LogScaleBackend,
}



from sigma.pipelines.common import (
    logsource_windows_ps_script,
    logsource_windows_driver_load,
    logsource_windows_dns_query,
    logsource_windows_image_load,
    logsource_windows_process_creation,
    logsource_linux_process_creation,
    logsource_windows_network_connection,
    logsource_windows_network_connection_initiated,
)
from sigma.processing.transformations import (
    ReplaceStringTransformation,
    AddConditionTransformation,
    ChangeLogsourceTransformation,
    DropDetectionItemTransformation,
    FieldMappingTransformation,
    DetectionItemFailureTransformation,
    MapStringTransformation,
)
from sigma.processing.conditions import (
    RuleContainsDetectionItemCondition,
    IncludeFieldCondition,
    MatchStringCondition,
)
from sigma.processing.pipeline import ProcessingItem, ProcessingPipeline
from sigma.processing.finalization import ConcatenateQueriesFinalizer


cond_field_parentbasefilename = IncludeFieldCondition(fields=["ParentBaseFileName"])
cond_field_contextbasefilename = IncludeFieldCondition(fields=["ContextBaseFileName"])


def generate_unsupported_process_field_processing_item(field):
    return ProcessingItem(
        identifier=f"cql_fail_process_start_{field}",
        transformation=DetectionItemFailureTransformation(
            f"Crowdstrike Query Language does not support the {field} field"
        ),
        rule_conditions=[logsource_windows_process_creation()],
        field_name_conditions=[IncludeFieldCondition(fields=[field])],
    )


def generate_unsupported_network_field_processing_item(field):
    return ProcessingItem(
        identifier=f"cql_fail_network_event_{field}",
        transformation=DetectionItemFailureTransformation(
            f"CrowdStrike Query Language does not support the {field} field"
        ),
        rule_conditions=[logsource_windows_network_connection()],
        field_name_conditions=[IncludeFieldCondition(fields=[field])],
    )


def generate_unsupported_dns_field_processing_item(field):
    return ProcessingItem(
        identifier=f"cql_fail_dns_event_{field}",
        transformation=DetectionItemFailureTransformation(
            f"CrowdStrike Query Language does not support the {field} field"
        ),
        rule_conditions=[logsource_windows_dns_query()],
        field_name_conditions=[IncludeFieldCondition(fields=[field])],
    )


def generate_unsupported_image_load_field_processing_item(field):
    return ProcessingItem(
        identifier=f"cql_fail_imageload_event_{field}",
        transformation=DetectionItemFailureTransformation(
            f"CrowdStrike Query Language does not support the {field} field"
        ),
        rule_conditions=[logsource_windows_image_load()],
        field_name_conditions=[IncludeFieldCondition(fields=[field])],
    )


def generate_unsupported_driverload_field_processing_item(field):
    return ProcessingItem(
        identifier=f"cql_fail_driverload_event_{field}",
        transformation=DetectionItemFailureTransformation(
            f"CrowdStrike Query Language does not support the {field} field"
        ),
        rule_conditions=[logsource_windows_driver_load()],
        field_name_conditions=[IncludeFieldCondition(fields=[field])],
    )

def common_processing_items():
    return [
       ProcessingItem(
        identifier="cql_win",
        transformation=AddConditionTransformation({"event_platform": "Win"}),
        rule_conditions=[
            logsource_windows_process_creation(),
            logsource_windows_dns_query(),
            logsource_windows_driver_load(),
            logsource_windows_image_load(),
            logsource_windows_ps_script(),
        ],
        rule_condition_linking=any,
        ),
        ProcessingItem(
            identifier="cql_process_creation_nix",
            transformation=AddConditionTransformation({"event_platform": "Lin"}),
            rule_conditions=[
                logsource_linux_process_creation(),
            ],
        ),
        ProcessingItem(
            identifier="cql_process_creation_fieldmaping",
            transformation=FieldMappingTransformation(
                {
                    "Image": "ImageFileName",
                    "ParentImage": "ParentBaseFileName",
                    "User": "UserName",
                    "ProcessId": "RawProcessId",
                    "sha256": "SHA256HashData",
                    "Computer": "ComputerName",
                    "OriginalFileName": "OriginalFilename",
                }
            ),
            rule_conditions=[
                logsource_windows_process_creation(),
                logsource_linux_process_creation(),
            ],
            rule_condition_linking=any,
        ),
        # Handle unsupported process start events
        generate_unsupported_process_field_processing_item("CurrentDirectory"),
        generate_unsupported_process_field_processing_item("imphash"),
        generate_unsupported_process_field_processing_item("md5"),
        generate_unsupported_process_field_processing_item("sha1"),
        generate_unsupported_process_field_processing_item("ParentCommandLine"),
        generate_unsupported_process_field_processing_item("FileVersion"),
        generate_unsupported_process_field_processing_item("Description"),
        generate_unsupported_process_field_processing_item("Product"),
        generate_unsupported_process_field_processing_item("Company"),
        generate_unsupported_process_field_processing_item("LogonGuid"),
        generate_unsupported_process_field_processing_item("ParentProcessGuid"),
        generate_unsupported_process_field_processing_item("ParentProcessId"),
        ProcessingItem(
            identifier="crowdstrike_process_creation_logsource",
            transformation=ChangeLogsourceTransformation(
                category="process_creation",
                product="windows",
                service="crowdstrike",
            ),
            rule_conditions=[
                logsource_windows_process_creation(),
            ],
        ),
        # ParentBaseFileName handling
        ProcessingItem(
            identifier="cql_parentbasefilename_fail_completepath",
            transformation=DetectionItemFailureTransformation(
                "Only file name of parent image is available in CrowdStrike Query Language events."
            ),
            field_name_conditions=[
                cond_field_parentbasefilename,
            ],
            detection_item_conditions=[
                MatchStringCondition(
                    cond="any",
                    pattern="^\\*\\\\?[^\\\\]+$",
                    negate=True,
                ),
            ],
        ),
        ProcessingItem(
            identifier="cql_parentbasefilename_executable_only",
            transformation=ReplaceStringTransformation(
                regex="^\\*\\\\([^\\\\]+)$",
                replacement="\\1",
            ),
            field_name_conditions=[
                cond_field_parentbasefilename,
            ],
        ),
        # IntegrityLevel handling
        ProcessingItem(
            identifier="cql_imagefilename_unknown_integrity_level",
            transformation=DetectionItemFailureTransformation(
                "Integrity Level needs to be one of Protected,System,High,Medium_Plus,Medium,Low,Untrusted"
            ),
            field_name_conditions=[
                IncludeFieldCondition(fields=["IntegrityLevel"])
            ],
            rule_conditions=[
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="Protected"
                ),
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="System"
                ),
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="High"
                ),
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="Medium_Plus"
                ),
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="Medium"
                ),
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="Low"
                ),
                RuleContainsDetectionItemCondition(
                    field="IntegrityLevel", value="Untrusted"
                ),
            ],
            rule_condition_linking=any,
            rule_condition_negation=True,
        ),
        ProcessingItem(
            identifier="cql_imagefilename_replace_integrity_level",
            transformation=MapStringTransformation(
                {
                    "Protected": "20480",
                    "System": "16384",
                    "High": "12288",
                    "Medium_Plus": "8448",
                    "Medium": "8192",
                    "Low": "4096",
                    "Untrusted": "0",
                }
            ),
            field_name_conditions=[
                IncludeFieldCondition(fields=["IntegrityLevel"])
            ],
        ),
        # Network Connection Common Processing Items
        ProcessingItem(
            identifier="cql_network_connection_fieldmapping",
            transformation=FieldMappingTransformation(
                {
                    "DestinationIp": "RemoteAddressIP4",
                    "DestinationPort": "RemotePort",
                    "Image": "ContextBaseFileName",
                    "SourcePort": "LocalPort",
                }
            ),
            rule_conditions=[
                logsource_windows_network_connection(),
            ],
        ),
        ProcessingItem(
            identifier="cql_network_connection_drop_initiated",
            transformation=DropDetectionItemTransformation(),
            rule_conditions=[
                logsource_windows_network_connection(),
            ],
            field_name_conditions=[
                IncludeFieldCondition(fields=["Initiated"]),
            ],
        ),
        ProcessingItem(
            identifier="crowdstrike_network_connection_logsource",
            transformation=ChangeLogsourceTransformation(
                category="network_connection",
                product="windows",
                service="crowdstrike",
            ),
            rule_conditions=[
                logsource_windows_network_connection(),
            ],
        ),
        generate_unsupported_network_field_processing_item('DetectionHostName'),
        # DNS Request Common Processing Items
        ProcessingItem(
            identifier="cql_dns_query_fieldmapping",
            transformation=FieldMappingTransformation(
                {
                    "QueryName": "DomainName",
                    "record_type": "RequestType",
                    "Image": "ContextBaseFileName",
                    "QueryResults": ["IP4Records", "IP6Records"],
                }
            ),
            rule_conditions=[
                logsource_windows_dns_query(),
            ],
        ),
        ProcessingItem(
            identifier="cql_dns_query_logsource",
            transformation=ChangeLogsourceTransformation(
                category="dns_query",
                product="windows",
                service="crowdstrike",
            ),
            rule_conditions=[
                logsource_windows_dns_query(),
            ],
        ),
        # Request Type handling
        ProcessingItem(
            identifier="cql_dns_unknown_request_type",
            transformation=DetectionItemFailureTransformation(
                "record_type needs to be one of A,NS,CNAME,PTR,MX,TXT,AAAA,ANY"
            ),
            field_name_conditions=[IncludeFieldCondition(fields=["RequestType"])],
            rule_conditions=[
                RuleContainsDetectionItemCondition(field="RequestType", value="A"),
                RuleContainsDetectionItemCondition(field="RequestType", value="NS"),
                RuleContainsDetectionItemCondition(
                    field="RequestType", value="CNAME"
                ),
                RuleContainsDetectionItemCondition(
                    field="RequestType", value="PTR"
                ),
                RuleContainsDetectionItemCondition(field="RequestType", value="MX"),
                RuleContainsDetectionItemCondition(
                    field="RequestType", value="TXT"
                ),
                RuleContainsDetectionItemCondition(
                    field="RequestType", value="AAAA"
                ),
                RuleContainsDetectionItemCondition(
                    field="RequestType", value="ANY"
                ),
            ],
            rule_condition_linking=any,
            rule_condition_negation=True,
        ),
        ProcessingItem(
            identifier="cql_dns_replace_request_type",
            transformation=MapStringTransformation(
                {
                    "A": "1",
                    "NS": "2",
                    "CNAME": "5",
                    "PTR": "12",
                    "MX": "15",
                    "TXT": "16",
                    "AAAA": "28",
                    "ANY": "255",
                }
            ),
            field_name_conditions=[IncludeFieldCondition(fields=["RequestType"])],
            field_name_condition_linking=any,
        ),
        # Query Results Handling
        # We wanna handle Query Results to always be evaluated as contains
        # This is because CrowdStrike returns the results in a semicolor seperated string
        ProcessingItem(
            identifier="cql_dns_wildcard_query_results",
            transformation=ReplaceStringTransformation(
                regex="(.*)", replacement="*\\1"
            ),
            field_name_conditions=[
                IncludeFieldCondition(fields="IP4Records"),
                IncludeFieldCondition(fields="IP6Records"),
            ],
            field_name_condition_linking=any,
            rule_conditions=[logsource_windows_dns_query()],
            rule_condition_linking=any,
        ),
        # Handle unsupported DNS query fields
        generate_unsupported_dns_field_processing_item("ProcessId"),
        generate_unsupported_dns_field_processing_item("QueryStatus"),
        generate_unsupported_dns_field_processing_item("answer"),
        # Image Load Common Processing Items
        ProcessingItem(
            identifier="cql_imageload_fieldmapping",
            transformation=FieldMappingTransformation(
                {
                    "Image": "TargetImageFileName",
                    "ImageLoaded": "ImageFileName",
                    "sha256": "SHA256HashData",
                    "md5": "MD5HashData",
                }
            ),
            rule_conditions=[
                logsource_windows_image_load(),
            ],
        ),
        generate_unsupported_image_load_field_processing_item("OriginalFileName"),
        generate_unsupported_image_load_field_processing_item("Description"),
        generate_unsupported_image_load_field_processing_item(
            "Signed"
        ),  # partially supported but not consistently
        generate_unsupported_image_load_field_processing_item("Imphash"),
        generate_unsupported_image_load_field_processing_item("CommandLine"),
        # Driver Load Common Processing Items
        ProcessingItem(
            identifier="cql_driverload_fieldmapping",
            transformation=FieldMappingTransformation(
                {
                    "ImageLoaded": "ImageFileName",
                    "sha256": "SHA256HashData",
                    "md5": "MD5HashData",
                    "OriginalFileName": "OriginalFilename",
                }
            ),
            rule_conditions=[
                logsource_windows_driver_load(),
            ],
        ),
        generate_unsupported_driverload_field_processing_item("Hashes"),
        generate_unsupported_driverload_field_processing_item("Imphash"),
        generate_unsupported_driverload_field_processing_item("Image"),
        # PowerShell Scripts Common Processing Items
        ProcessingItem(
            identifier="cql_powershell_script_fieldmapping",
            transformation=FieldMappingTransformation(
                {
                    "Payload": ["CommandHistory", "ScriptContent"],
                    "ScriptBlockText": ["CommandHistory", "ScriptContent"],
                }
            ),
            rule_conditions=[logsource_windows_ps_script()],
            rule_condition_linking=any,
        ),
        # ContextBaseFileName handling
        ProcessingItem(
            identifier="cql_contextbasefilename_fail_completepath",
            transformation=DetectionItemFailureTransformation(
                "Only file name of image is available in CrowdStrike Query Language events."
            ),
            field_name_conditions=[
                cond_field_contextbasefilename,
            ],
            detection_item_conditions=[
                MatchStringCondition(
                    cond="any",
                    pattern="^\\*\\\\?[^\\\\]+$",
                    negate=True,
                ),
            ],
        ),
        ProcessingItem(
            identifier="cql_contextbasefilename_executable_only",
            transformation=ReplaceStringTransformation(
                regex="^\\*\\\\([^\\\\]+)$",
                replacement="\\1",
            ),
            field_name_conditions=[
                cond_field_contextbasefilename,
            ],
        ),
        # ImageFileName full path handling
        ProcessingItem(
            identifier="cql_imagefilename_replace_disk_name",
            transformation=ReplaceStringTransformation(
                regex="[C-Z]:", replacement="\\\\Device\\\\HarddiskVolume?", skip_special=True, interpret_special=True
            ),
            field_name_conditions=[
                IncludeFieldCondition(fields=["ImageFileName"]),
                IncludeFieldCondition(fields=["TargetImageFileName"]),
            ],
            field_name_condition_linking=any,
        ),
        ProcessingItem(
            identifier="cql_imagefilename_replace_disk_name",
            transformation=ReplaceStringTransformation(regex=":", replacement="", skip_special=True),
            field_name_conditions=[
                IncludeFieldCondition(fields=["ImageFileName"]),
                IncludeFieldCondition(fields=["TargetImageFileName"]),
            ],
            field_name_condition_linking=any,
        ),
    ]

def crowdstrike_fdr_pipeline() -> ProcessingPipeline:
    return ProcessingPipeline(
        name="Generic Log Sources to CrowdStrike Falcon Data Replicator (FDR) Transformation",
        priority=10,
        items=[
            # Process Creation
            ProcessingItem(
                identifier="cs_process_creation_eventtype",
                transformation=AddConditionTransformation(
                    {
                        "event_simpleName": [
                            "ProcessRollup2",
                            "SyntheticProcessRollup2",
                        ],
                    }
                ),
                rule_conditions=[
                    logsource_windows_process_creation(),
                ],
            ),
            # DNS Request
            ProcessingItem(
                identifier="cql_dns_query_eventtype",
                transformation=AddConditionTransformation(
                    {
                        "event_simpleName": "DnsRequest",
                    }
                ),
                rule_conditions=[logsource_windows_dns_query()],
            ),
            # Network Connections
            ProcessingItem(
                identifier="cql_network_connection_eventtype_connect",
                transformation=AddConditionTransformation(
                    {"event_simpleName": "NetworkConnectIP4"}
                ),
                rule_conditions=[
                    logsource_windows_network_connection(),
                    logsource_windows_network_connection_initiated(True),
                ],
            ),
            ProcessingItem(
                identifier="cql_network_connection_eventtype_accept",
                transformation=AddConditionTransformation(
                    {
                        "event_simpleName": "NetworkReceiveAcceptIP4",
                    }
                ),
                rule_conditions=[
                    logsource_windows_network_connection(),
                    logsource_windows_network_connection_initiated(False),
                ],
            ),
            # Driver Load
            ProcessingItem(
                identifier="cql_driverload_eventtype",
                transformation=AddConditionTransformation(
                    {"event_simpleName": "DriverLoad"}
                ),
                rule_conditions=[logsource_windows_driver_load()],
            ),
            # Image Load
            ProcessingItem(
                identifier="cql_imageload_eventtype",
                transformation=AddConditionTransformation(
                    {"event_simpleName": "ClassifiedModuleLoad"}
                ),
                rule_conditions=[logsource_windows_image_load()],
                rule_condition_linking=any,
            ),
            # Powershell Scripting
            ProcessingItem(
                identifier="cql_powershell_script_eventtype",
                transformation=AddConditionTransformation(
                    {
                        "event_simpleName": [
                            "CommandHistory",
                            "ScriptControlScanTelemetry",
                        ]
                    }
                ),
                rule_conditions=[
                    logsource_windows_ps_script(),
                ],
            ),
        ] + common_processing_items(),
    )

def crowdstrike_falcon_pipeline() -> ProcessingPipeline:
    return ProcessingPipeline(
        name="CrowdStrike Falcon Pipeline",
        priority=10,
        items=[
            # Process Creation
            ProcessingItem(
                identifier="cql_process_creation_eventtype",
                transformation=AddConditionTransformation(
                    {"#event_simpleName": ["ProcessRollup2", "SyntheticProcessRollup2"]}
                ),
                rule_conditions=[
                    logsource_windows_process_creation(),
                    logsource_linux_process_creation(),
                ],
                rule_condition_linking=any,
            ),
            # DNS Request
            ProcessingItem(
                identifier="cql_dns_query_eventtype",
                transformation=AddConditionTransformation(
                    {
                        "#event_simpleName": "DnsRequest",
                    }
                ),
                rule_conditions=[logsource_windows_dns_query()],
            ),
            # Network Connections
            ProcessingItem(
                identifier="cql_network_connection_eventtype_connect",
                transformation=AddConditionTransformation(
                    {"#event_simpleName": "NetworkConnectIP4"}
                ),
                rule_conditions=[
                    logsource_windows_network_connection(),
                    logsource_windows_network_connection_initiated(True),
                ],
            ),
            ProcessingItem(
                identifier="cql_network_connection_eventtype_accept",
                transformation=AddConditionTransformation(
                    {
                        "#event_simpleName": "NetworkReceiveAcceptIP4",
                    }
                ),
                rule_conditions=[
                    logsource_windows_network_connection(),
                    logsource_windows_network_connection_initiated(False),
                ],
            ),
            # Driver Load
            ProcessingItem(
                identifier="cql_driverload_eventtype",
                transformation=AddConditionTransformation(
                    {"#event_simpleName": "DriverLoad"}
                ),
                rule_conditions=[logsource_windows_driver_load()],
            ),
            # Image Load
            ProcessingItem(
                identifier="cql_imageload_eventtype",
                transformation=AddConditionTransformation(
                    {"#event_simpleName": "ClassifiedModuleLoad"}
                ),
                rule_conditions=[logsource_windows_image_load()],
                rule_condition_linking=any,
            ),
            # Powershell Scripting
            ProcessingItem(
                identifier="cql_powershell_script_eventtype",
                transformation=AddConditionTransformation(
                    {
                        "#event_simpleName": [
                            "CommandHistory",
                            "ScriptControlScanTelemetry",
                        ]
                    }
                ),
                rule_conditions=[
                    logsource_windows_ps_script(),
                ],
            ), 
        ] + common_processing_items(),
        finalizers=[ConcatenateQueriesFinalizer()],
    )
